import enum
import functools

from vazio.protocol.multigauge import encode_request, decode_reply, Command, Channel


class Enum(enum.Enum):
    @classmethod
    def encode(cls, value):
        return cls(value).value


class IntFlag(enum.IntFlag):
    @classmethod
    def decode(cls, data):
        return cls(ord(data))


class Remote(Enum):
    Local = "0"
    Remote = "1"
    Serial = "2"


class HighVoltage(Enum):
    Off = "0"
    OnStartStep = "1"
    On = "1"
    OnStartFixed = "2"
    OnProtectStep = "3"
    OnProtectFixed = "4"
    OffPanelInterlock = "-3"
    OffRemoteInterlock = "-4"
    OffHVOverTemperature = "-8"
    OffRemoteFault = "-7"
    OffHVProtect = "-6"
    OffHVShortCircuit = "-5"


class Unit(Enum):
    torr = "0"
    mbar = "1"
    pascal = "2"


class HVDeviceNumber(Enum):
    Spare = "0"
    SCTr_500 = "1"
    SCTr_300 = "2"
    SCTr_150 = "3"
    SCTr_75_55_40 = "4"
    SCTr_20 = "4"
    DiodeND_500 = "6"
    DiodeND_300 = "7"
    DiodeND_150 = "8"
    DiodeND_75_55_40 = "9"
    DiodeND_20 = ":"


class GaugeDeviceNumber(Enum):
    Convector = "0"
    MiniBA = "1"
    ColdCathode = "2"


class SerialDeviceNumber(Enum):
    RS232 = "0"
    RS485 = "1"


class InterlockStatus(IntFlag):
    FrontPanel = 2
    HV1Remote = 4
    HV1Cable = 8
    HV2Remote = 64
    HV2Cable = 128


class FixedStep(Enum):
    Fixed = "0"
    Step = "1"


class StartProtect(Enum):
    Start = "0"
    Protect = "1"


class Polarity(Enum):
    Negative = "0"
    Positive = "1"


class RemoteOutput(IntFlag):
    HighVoltageEnable = 1
    SetPoint2Active = 2
    SetPoint1Active = 4
    InterlockActive = 8
    HighVoltageFault = 16
    SerialMode = 32
    ProtectMode = 64


class RemoteInput(IntFlag):
    StepMode = 4
    RemoteMode = 8
    ProtectMode = 16
    HVOutputEnable = 32
    HVConfirm = 64
    RemoteInterlock = 128


ProtocolErrors = {
    # If incongruencies are detected in the composition of the
    # data packet sent to the Dual controller (in other words a
    # correct reception but an incorrect data format), the Dual
    # controller will reply with an error code identified by the
    # .!. (21h) command according to the following table.
    "1": "Reserved (checksum error)",
    "2": "Non existent command code",
    "3": "Channel not valid for the selected command",
    "4": "Write mode not allowed for the selected command",
    "5": "Unvalid or non-congruent data transmitted",
    "6": "Write value exceeding the allowed limits or step not allowed",
    "7": "Data format not recognized on the protocols implemented",
    "8": "Write not allowed to channel ON",
    "9": "Write not allowed to channel OFF",
    ":": "Write allowed in Serial Configuration Mode only",
}


ControllerErrors = {
    # Value  Error type  Error reference
    # High Voltage errors
    "HV": {
        "1": "High Voltage off due to front panel interlock activation Panel Interlock",
        "2": "High Voltage off due to Remote I/O interlock activation Remote Interlock",
        "3": "High Voltage off due to Cable HV interlock activation Cable Interlock",
        "4": "Dual fault HV not found",
        "5": "High Voltage off due to a general DSP determined fault HV Fault",
        "6": "High Voltage off due to an HV module overtemperature determined by the DSP HV Overtemperature",
        "7": "Remote I/O card not present or faulty R.I/O not found",
        "8": "Remote I/O card present, but faulty R.I/O fault",
        "9": "High Voltage off due to the protect function activation Protect",
        "10": "High Voltage off due to shortcircuit protection activation Short Circuit",
        "11": "High Voltage off due to an HV module overvoltage or overcurrent determined by the DSP Over Volt/Curr",
        "12": "High Voltage off due to the zero measurement protection activation Zero Meas MiniGauge errors ",
    },
    "MG": {
        "1": "MiniGauge off due to front panel interlock activation Panel Interlock",
        "2": "The selected Minigauge was not recognized Gauge not found",
        "3": "The Minigauge is signaling a Fault condition Gauge fault",
        "4": "The selected Minigauge was disconnected Gauge not connected System errors ",
    },
    "SW": {
        "1": "RAM failure: RAM diagnostics error Software Error",
        "2": "config register: incorrect value in the uC 68HC11 configuration register Software Error",
        "3": "test mode: invalid uC 68HC11 operating mode Software Error",
        "4": "copyright: violation of the signature in the ROM or the ROM was corrupted Software Error",
        "5": "eeprom fault: checksum or non-volatile memory write errors. Factory defaults are automatically loaded Software Error",
        "6": "version number: incompatible uC and Dsp versions Software Error",
        "7": "hv dsp not found: the Dsp does not respond during the uC initialization phase Software Error",
        "8": "dsp fault: the Dsp does not respond during normal operation Software Error ",
        "9": "invalid option: option card not configured correctly Software Error ",
        "10": "unknow option: generic execution error Software Error",
    },
}


def nop(v):
    return v


class Value:
    def __init__(self, command, decode=nop, encode=None):
        self.command = command
        self.decode = decode
        self.encode = encode

    def _read(self, ctrl, channel):
        request = encode_request(channel, self.command, "?")
        reply = ctrl.conn.write_readline(request)
        *_, value = decode_reply(reply)
        return self.decode(value)

    def _write(self, ctrl, channel, value):
        if self.encode is None:
            raise AttributeError(
                "can't set: {.name} on {.name}".format(self.command, channel)
            )
        request = encode_request(channel, self.command, self.encode(value))
        ctrl.conn.write_readline(request)

    def __get__(self, ctrl, owner=None):
        return self._read(ctrl, Channel.NoChannel)

    def __set__(self, ctrl, value):
        self._write(ctrl, Channel.NoChannel, value)


class ChannelValue(Value):
    def __get__(self, channel, owner=None):
        return self._read(channel.ctrl, channel.channel)

    def __set__(self, channel, value):
        self._write(channel.ctrl, channel.channel, value)


class BaseChannel:

    device_type = ChannelValue(Command.DeviceType)
    error_status = ChannelValue(Command.ErrorStatus, decode=ControllerErrors.get)

    def __init__(self, channel, ctrl=None):
        self.channel = channel
        self.ctrl = ctrl

    def __get__(self, ctrl, owner=None):
        ch = ctrl._channels.get(self.channel)
        if ch is None:
            ch = type(self)(self.channel, ctrl)
            ctrl._channels[self.channel] = ch
        return ch


def EnumCV(cmd, enu, **kwargs):
    kwargs.setdefault("decode", enu)
    if hasattr(enu, "encode"):
        kwargs.setdefault("encode", enu.encode)
    return ChannelValue(cmd, **kwargs)


IntCV = functools.partial(ChannelValue, encode=int, decode=str)
IntCVRO = functools.partial(ChannelValue, decode=int)
FloatCV = functools.partial(ChannelValue, decode=float, encode=str)
FloatCVRO = functools.partial(ChannelValue, decode=float)


class HV(BaseChannel):

    high_voltage = EnumCV(Command.HighVoltage, HighVoltage)
    device_number = EnumCV(Command.DeviceNumber, HVDeviceNumber)

    voltage = IntCVRO(Command.Voltage)
    current = FloatCVRO(Command.Current)
    pressure = FloatCVRO(Command.Pressure)

    fixed_step = EnumCV(Command.FixedStep, FixedStep)
    start_protect = EnumCV(Command.StartProtect, StartProtect)
    polarity = EnumCV(Command.Polarity, Polarity)

    voltage_max = IntCV(Command.VoltageMax)  # [3000, 7000] step 100 (V)
    current_max = IntCV(Command.CurrentMax)  # [100, 400] step 10 (mA)
    power_max = IntCV(Command.PowerMax)  # [100, 400] step 10 (W)
    current_protect = IntCV(Command.CurrentProtect)  # [10, 100] step 10 (mA)
    voltage_step1 = IntCV(Command.VoltageStep1)  # [3000, 7000] step 100 (V)
    current_step1 = FloatCV(Command.CurrentStep1)  # [1e-9, 1e1] (A)
    voltage_step2 = IntCV(Command.VoltageStep2)  # [3000, 7000] step 100 (V)
    current_step2 = FloatCV(Command.CurrentStep2)  # [1e-9, 1e1] (A)
    set_point1 = FloatCV(Command.SetPoint1)  # [1e-9, 1e1] (torr) (> sp2)
    set_point2 = FloatCV(Command.SetPoint2)  # [1e-9, 1e1] (torr)
    remote_output = ChannelValue(Command.RemoteOutput, decode=RemoteOutput.decode)
    remote_input = ChannelValue(Command.RemoteInput, decode=RemoteInput.decode)


class Gauge(BaseChannel):

    device_number = ChannelValue(Command.DeviceNumber, decode=GaugeDeviceNumber)
    pressure = ChannelValue(Command.Pressure, decode=float)


class Serial(BaseChannel):

    device_number = ChannelValue(Command.DeviceNumber, decode=SerialDeviceNumber)


class VarianDual:
    """
    VarianDual controller based on MultiGauge protocol

    conn: any object with write_readline (should be configured
          with eol='\r')
    """

    remote = Value(Command.Remote, decode=Remote, encode=Remote.encode)
    unit = Value(Command.Unit, decode=Unit, encode=Unit.encode)

    error_status = Value(Command.ErrorStatus, decode=ControllerErrors.get)
    interlock_status = Value(Command.InterlockStatus, decode=InterlockStatus.decode)
    ctrl_firmware_version = Value(Command.MicroControllerFirmwareVersion)
    dsp_firmware_version = Value(Command.DSPFirmwareVersion)

    hv1 = HV(Channel.HighVoltage1)
    hv2 = HV(Channel.HighVoltage2)
    gauge1 = Gauge(Channel.Gauge1)
    gauge2 = Gauge(Channel.Gauge2)
    serial = Serial(Channel.Serial)

    serial_config = Value(
        Command.SerialConfig,
        decode=lambda v: v == "1",
        encode=lambda v: "1" if v else "0",
    )

    def __init__(self, conn):
        self.conn = conn
        self._channels = {}

        # TODO:
        # on connect:
        # - set ACK reply mode
        # - set unit to mbar



"""
MultiGauge protocol

header command = '#'
header response = '>'
terminator = '\r'

Host to controller command format:
[header command(1)] [channel(1)] [command(2)] [data] [terminator(1)]

Controller to Host reply format:
[header response(1)] [channel(1)] [command(2)] [data] [terminator(1)]
"""

import enum


HEADER_REQ = "#"
HEADER_REP = ">"
ACK = '\x06'


class Enum(enum.Enum):

    @staticmethod
    def _size():
        return 1

    @classmethod
    def decode(cls, data):
        if isinstance(data, bytes):
            data = data.decode()
        elif isinstance(data, int):
            data = '{{:0{}}}'.format(cls._size()).format(data)
        return cls(data)


class Channel(Enum):
    NoChannel = "0"
    HighVoltage1 = "1"
    HighVoltage2 = "2"
    Gauge1 = "3"
    Gauge2 = "4"
    Serial = "5"


class Command(Enum):
    Remote = "10"
    HighVoltage = "30"
    Unit = "03"
    MicroControllerFirmwareVersion = "05"
    DSPFirmwareVersion = "04"
    DeviceNumber = "01"
    DeviceType = "11"
    Voltage = "07"
    Current = "08"
    Pressure = "02"
    ErrorStatus = "19"
    SerialReset = "06"
    RemoteError = "12"
    InterlockStatus = "13"
    FixedStep = "60"
    StartProtect = "61"
    Polarity = "62"
    VoltageMax = "63"
    CurrentMax = "64"
    PowerMax = "65"
    CurrentProtect = "66"
    VoltageStep1 = "67"
    CurrentStep1 = "68"
    VoltageStep2 = "69"
    CurrentStep2 = "70"
    SetPoint1 = "71"
    SetPoint2 = "72"
    RemoteOutput = "73"
    RemoteInput = "74"
    SerialConfig = "80"

    @staticmethod
    def _size():
        return 2


def encode(header, channel, command, data):
    channel = Channel.decode(channel).value
    command = Command.decode(command).value
    if isinstance(data, bytes):
        data = data.decode()
    if isinstance(data, enum.Enum):
        data = data.value
    return '{}{}{}{}\r'.format(header, channel, command, data).encode()


def encode_request(channel, command, data):
    return encode(HEADER_REQ, channel, command, data)


def encode_reply(channel, command, data):
    return encode(HEADER_REP, channel, command, data)


def decode(data):
    data = data.decode()
    assert data[-1] == '\r'
    return (
        data[0],
        Channel.decode(data[1:2]),
        Command.decode(data[2:4]),
        data[4:-1]
    )


def decode_reply(data):
    result = decode(data)
    assert result[0] == HEADER_REP
    return result



"""
MKS simulator helper classes

To create a MKS device use the following configuration as
a starting point:

.. code-block:: yaml

    devices:
    - class: MKS937
      name: mks-1
      transports:
      - type: serial
        url: /tmp/mks00
"""

import random

from sinstruments.simulator import BaseDevice


def funiform(a, b):
    return lambda: '{:7.1E}'.format(random.uniform(a, b))


def iuniform(a, b):
    return lambda: '{:05d}'.format(random.randint(a, b))


state = {
    'P1': funiform(5e-9, 9e-7),
    'P2': funiform(5e-9, 9e-3),
    'P3': funiform(5e-9, 9e-7),
    'P4': funiform(5e-9, 9e-3),
    'P5': funiform(5e-9, 9e-7),
    'C1': funiform(5e-9, 9e-3),
    'C2': funiform(5e-9, 9e-7),
    'C3': funiform(5e-9, 9e-3),
    'C4': funiform(5e-9, 9e-7),
    'C5': funiform(5e-9, 9e-3),
    'GAUGES': 'CvPrCv',
    'PRO1': '4.1E-9',
    'PRO2': '5.1E-9',
    'PRO3': '6.1E-9',
    'PRO4': '7.1E-9',
    'PRO5': '8.1E-9',
    'RELAYS': 'rly11011',
    'RLY1': '4.7E-9',
    'RLY2': '5.7E-9',
    'RLY3': '6.7E-9',
    'RLY4': '7.7E-9',
    'RLY5': '8.7E-9',
    'VER': '2.59,6.17',
    'UNIT': 'mbar',
    'ECC1': 'OK',
    'ECC2': 'PROTECT!',
    'ECC4': 'PS_FAULT!',
    'XCC1': 'OK',
    'XCC2': 'OK',
    'XCC4': 'OK',
}


class MKS937(BaseDevice):

    newline = b"\r"
    baudrate = 19200   # should accept 19200 or lower

    def handle_message(self, line):
        self._log.info("processing message %r", line)
        line = line.decode()
        data = state[line]
        if callable(data):
            data = data()
        reply = data.encode() + self.newline
        self._log.debug("reply %r", reply)
        return reply



"""
Agilent 4UHV simulator helper classes

To create an Agilent 4UHV device use the following configuration as
a starting point:

.. code-block:: yaml

    devices:
    - class: Agilent4UHV
      name: agilent-1
      transports:
      - type: serial
        url: /tmp/agilent4uhv00
"""

import enum
import random
import operator
import functools

from sinstruments.simulator import BaseDevice, MessageProtocol


def funiform(a, b):
    return lambda: '{: 10.1E}'.format(random.uniform(a, b))


def iuniform(a, b, x=1):
    return lambda: '{:05d}'.format(x*random.randint(a, b))


state = {
}

# protocol
# MESSAGE: <STX>+<ADDR>+<WIN>+<COM>+<DATA>+<ETX>+<CRC>

STX = b"\x02"
ETX = b"\x03"
ACK = b"\x06"
NACK = b"\x15"
UNKNOWN_WINDOW = b"\x32"
DATA_TYPE_ERROR = b"\x33"
WINDOW_DISABLED = b"\x35"
RS232_ADDR = b"\x80"


class Command(enum.Enum):
    READ = b"\x30"
    WRITE = b"\31"


class Window(enum.Enum):
    LocalMode = b"008"
    HV1_ON = b"011"
    HV2_ON = b"012"
    HV3_ON = b"013"
    HV4_ON = b"014"
    Status = b"205"
    ErrorCode = b"206"
    Model = b"319"

    SerialNumber = b"323"
    SerialType = b"504"
    Channel = b"505"
    Unit = b"600"
    Mode = b"601"
    Protect = b"602"
    FixedStep = b"603"

    Dev1 = b"610"
    Power1 = b"612"
    Vt1 = b"613"
    IProt1 = b"614"
    SP1 = b"615"

    Dev2 = b"620"
    Power2 = b"622"
    Vt2 = b"623"
    IProt2 = b"624"
    SP2 = b"625"

    Dev3 = b"630"
    Power3 = b"632"
    Vt3 = b"633"
    IProt3 = b"634"
    SP3 = b"635"

    Dev4 = b"640"
    Power4 = b"642"
    Vt4 = b"643"
    IProt4 = b"644"
    SP4 = b"645"

    Temp1 = b"801"
    Temp2 = b"802"
    Temp3 = b"808"
    Temp4 = b"809"

    Ilock = b"803"
    StatusSP = b"804"

    V1 = b"810"
    I1 = b"811"
    P1 = b"812"

    V2 = b"820"
    I2 = b"821"
    P2 = b"822"

    V3 = b"830"
    I3 = b"831"
    P3 = b"832"

    V4 = b"840"
    I4 = b"841"
    P4 = b"842"


crc = functools.partial(functools.reduce, operator.xor)


def crc_ascii(msg):
    return "{:X}".format(crc(msg)).encode()


def decode_message(msg):
    assert msg[0:1] == STX, "invalid start byte"
    addr = msg[1] - RS232_ADDR[0]
    wnd = Window(msg[2:5])
    cmd = Command(msg[5:6])
    data = msg[5:-3]
    assert msg[-3:-2] == ETX, "invalid end byte"
    assert crc_ascii(msg[1:-2]) == msg[-2:], "invalid crc"
    return addr, wnd, cmd, data


def encode_answer(wnd, data=ACK, addr=0):
    if isinstance(data, str):
        data = data.encode()
    addr = chr(ord(RS232_ADDR) + addr).encode()
    msg = STX + addr + wnd.value + data + ETX
    msg += crc_ascii(msg[1:])
    return msg


def read_messages(channel):
    def stream():
        while True:
            yield channel.read1()

    buff = b""
    for data in stream():
        buff += data
        messages = buff.split(STX)
        for msg in messages[1:-1]:
            yield STX + msg
        last = STX + messages[-1]
        if len(last) > 2 and last[-3:-2] == ETX:
            buff = b""
            yield last
        else:
            buff = last


class WindowProtocol(MessageProtocol):

    def read_messages(self):
        return read_messages(self.channel)


state = {
    Window.LocalMode: "000001",      # SERIAL:000000, REMOTE:000001, LOCAL:000002
    Window.ErrorCode: b"000000",
    Window.Model: b"4UHV",
    Window.P1: funiform(5e-9, 9e-7),
    Window.P2: funiform(5e-9, 9e-7),
    Window.P3: funiform(5e-9, 9e-7),
    Window.P4: funiform(5e-9, 9e-7),
    Window.I1: funiform(5e-3, 9e-2),
    Window.I2: funiform(5e-3, 9e-2),
    Window.I3: funiform(5e-3, 9e-2),
    Window.I4: funiform(5e-3, 9e-2),
    Window.V1: iuniform(5, 68, 100),
    Window.V2: iuniform(50, 68, 100),
    Window.V3: iuniform(50, 68, 100),
    Window.V4: iuniform(50, 68, 100),
}


class Agilent4UHV(BaseDevice):

    protocol = WindowProtocol
    baudrate = 9600   # should accept 9600 or lower

    def handle_message(self, line):
        self._log.info("processing message %r", line)
        addr, wnd, cmd, data = decode_message(line)
        if cmd == Command.READ:
            data = state[wnd]
            if callable(data):
                data = data()
            reply = encode_answer(wnd, data)
            self._log.info("reply with %r", reply)
            return reply
